"""
Transparently load variables from environment or JSON/YAML file.
"""
import json
import logging
import os

from decimal import Decimal
from distutils.util import strtobool
from io import StringIO
from typing import TypeVar, List, Callable

log = logging.getLogger(__name__)

CASTS = [int, str, float, list, bool, Decimal]
CastTypes = TypeVar('CastTypes', *CASTS)


class RequiredValueMissing(Exception):
    pass


def _load_config(path: str) -> dict:
    """
    Given a file path, parse it based on its extension (YAML or JSON)
    and return the values as a Python dictionary. JSON is the default if an
    extension can't be determined.
    """
    __, ext = os.path.splitext(path)
    if ext in ['.yaml', '.yml']:
        import ruamel.yaml
        loader = ruamel.yaml.safe_load
    else:
        loader = json.load
    with open(path) as f:
        config = loader(f)
    log.info("Loaded config from %s", path)
    return config


def _find_file(file: str, verify_later: bool = False) -> str:
    if os.path.isabs(file):
        config_file = file
    else:
        config_file = os.path.join(os.getcwd(), file)
    if not verify_later and not os.path.exists(config_file):
        return None
    return config_file


class Value:
    def __init__(self, key: str, default=None, required: bool = None,
                 initial: Callable[[], CastTypes] = None,
                 cast_as: CastTypes = None, help: str = ""):
        """

        :param key:      Name of the value used in file or environment variable
        :param default:  Default value if none is provided.
        :param required: Loading a config will fail if a value is not provided.
                         Defaults to True if no default is provided otherwise
                         False.
        :param initial:  Initial value to use when generating a config
        :param cast_as:  Python type to cast variable as. Defaults to type of
                         default (if provided) or str.
        :param help:     Plain-text description of the value.
        """
        self.key = key
        self.default = default
        if default is not None:
            self.value = default
        self.required = bool(default is None or required)
        if initial and not callable(initial):
            raise ValueError("Initial value must be a callable.")
        self._initial = initial
        self.help = help
        if cast_as:
            self.cast_as = cast_as
        elif default is not None:
            self.cast_as = type(default)
        else:
            self.cast_as = str

    @property
    def initial(self):
        if callable(self._initial):
            return self._initial()
        elif self.default is not None:
            return self.default
        return ''

    def set_value(self, defined: CastTypes = None):
        if self.key in os.environ:
            self.value = self.cast(os.environ[self.key])
        elif defined is not None:
            assert(type(defined) == self.cast_as)
            self.value = defined
        elif self.required:
            raise RequiredValueMissing(self.key)

    def cast(self, val: str):
        """converts string to type requested by `cast_as`"""
        try:
            return getattr(self, 'cast_as_{}'.format(
                self.cast_as.__name__.lower()))(val)
        except AttributeError:
            return self.cast_as(val)

    def cast_as_list(self, val: str) -> list:
        """Convert a comma-separated string to a list"""
        return val.split(',')

    def cast_as_bool(self, val: str) -> bool:
        """
        True values are y, yes, t, true, on and 1
        False values are n, no, f, false, off and 0
        Raises ValueError if val is anything else.
        """
        return bool(strtobool(val))


class GoodConf:
    def __init__(self, description: str = '',
                 file_env_var: str = None,
                 default_files: List[str] = None):
        """

        :param description: a plain-text description used as a header when
                            generating the file
        :param file_env_var: the name of an environment variable which can be
                             used for the name of the configuration file to
                             load
        :param default_files: if no file is given, try to load a configuration
                              from these files in order
        """
        self.description = description
        self.file_env_var = file_env_var
        self.config_file = None
        self.default_files = default_files
        self._values = {}

    def __getattr__(self, key: str):
        return self._values[key].value

    def load(self, file: str = None):
        """Find config file and set values"""
        self.config_file = self.determine_file(file)
        if self.config_file:
            config = _load_config(self.config_file)
        else:
            config = {}
        self.set_values(config)

    def determine_file(self, file: str = None):
        """
        Return absolute path to the config file or None if it does not exist.
        Relative paths will be resolved relative to the working directory
        Will return the first file of:
        1. os.environ[self.config_file_env_var] (if defined)
        2. config file passed as arg
        3. first default file found
        """
        if self.file_env_var and self.file_env_var in os.environ:
            return _find_file(os.environ[self.file_env_var],
                              verify_later=True)
        if file:
            return _find_file(file, verify_later=True)
        if self.default_files:
            for f in self.default_files:
                default_file = _find_file(f)
                if default_file:
                    return default_file

    def define_values(self, *args: Value):
        """Sets up internal dict used to track values"""
        for val in args:
            self._values[val.key] = val

    def set_values(self, config: dict):
        for k, v in self._values.items():
            v.set_value(config.get(k))

    def get_initial(self):
        return {k: v.initial for k, v in self._values.items()}

    def generate_yaml(self):
        """
        Dumps initial config in YAML
        """
        import ruamel.yaml
        yaml = ruamel.yaml.YAML()
        yaml_str = StringIO()
        yaml.dump(self.get_initial(), stream=yaml_str)
        yaml_str.seek(0)
        dict_from_yaml = yaml.load(yaml_str)
        if self.description:
            dict_from_yaml.yaml_set_start_comment(
                '\n' + self.description + '\n\n')
        for k in dict_from_yaml.keys():
            if self._values[k].help:
                dict_from_yaml.yaml_set_comment_before_after_key(
                    k, before='\n' + self._values[k].help)
        yaml_str = StringIO()
        yaml.dump(dict_from_yaml, yaml_str)
        yaml_str.seek(0)
        return yaml_str.read()

    def generate_json(self):
        """
        Dumps initial config in JSON
        """
        return json.dumps(self.get_initial(), indent=2)

    def generate_markdown(self):
        """
        Documents values in markdown
        """
        lines = []
        if self.description:
            lines.extend(['# {}'.format(self.description), ''])
        for k, v in self._values.items():
            lines.append('* **{}**  '.format(k))
            if v.required:
                lines[-1] = lines[-1] + '_REQUIRED_  '
            if v.help:
                lines.append('  {}  '.format(v.help))
            lines.append('  type: `{}`  '.format(v.cast_as.__name__))
            if v.default:
                lines.append('  default: `{}`  '.format(v.default))
        return '\n'.join(lines)



import functools
import os
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from unittest import mock, TestCase, SkipTest

from goodconf import (Value, GoodConf, RequiredValueMissing,
                      _load_config, _find_file)
from goodconf.contrib.django import execute_from_command_line_with_config

KEY = 'GOODCONF_TEST'


@contextmanager
def env_var(key, value):
    os.environ[key] = value
    yield
    del(os.environ[key])


class ValueTests(TestCase):
    def test_default_not_required(self):
        """Values with a default are not required"""
        v = Value(KEY, default='s')
        self.assertFalse(v.required)

    def test_no_default_required(self):
        """Values with a default are not required"""
        v = Value(KEY)
        self.assertTrue(v.required, True)

    def test_infer_cast(self):
        """If not provided, cast is type of default"""
        v = Value(KEY, default=5)
        self.assertEqual(v.cast_as, int)

    def test_default_cast(self):
        """Default cast is str"""
        v = Value(KEY)
        self.assertEqual(v.cast_as, str)

    def test_explicit_cast(self):
        v = Value(KEY, cast_as=bool)
        self.assertEqual(v.cast_as, bool)

    def test_cast_bool(self):
        v = Value(KEY, cast_as=bool)
        with env_var(KEY, 'true'):
            v.set_value()
            self.assertTrue(v.value)
        with env_var(KEY, 'false'):
            v.set_value()
            self.assertFalse(v.value)

    def test_initial(self):
        v = Value(KEY, initial=lambda: 'x')
        self.assertEqual(v.initial, 'x')

    def test_initial_bad(self):
        self.assertRaises(ValueError, Value, KEY, initial='x')

    def test_intial_default(self):
        v = Value(KEY, default='x')
        self.assertEqual(v.initial, 'x')

    def test_no_initial(self):
        v = Value(KEY)
        self.assertEqual(v.initial, '')

    def test_defined_value(self):
        v = Value(KEY)
        v.set_value('x')
        self.assertEqual(v.value, 'x')

    def test_env_var_precedence(self):
        v = Value(KEY)
        with env_var(KEY, 'y'):
            v.set_value('x')
        self.assertEqual(v.value, 'y')

    def test_required_no_value(self):
        v = Value(KEY)
        self.assertRaises(RequiredValueMissing, v.set_value)

    def test_cast_list(self):
        v = Value(KEY, cast_as=list)
        with env_var(KEY, 'a,b,c'):
            v.set_value()
        self.assertListEqual(v.value, ['a', 'b', 'c'])


def skip_if_no_yaml(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            import ruamel.yaml
            return f(*args, **kwargs)
        except ImportError:
            raise SkipTest("[yaml] extras is not installed")
    return wrapper


class TestFileHelpers(TestCase):
    def setUp(self):
        self.tmpdir = TemporaryDirectory()

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_json(self):
        conf = os.path.join(self.tmpdir.name, 'conf.json')
        with open(conf, 'w') as f:
            f.write('{"a": "b", "c": 3}')
        self.assertEqual({'a': 'b', 'c': 3}, _load_config(conf))

    @skip_if_no_yaml
    def test_yaml(self):
        conf = os.path.join(self.tmpdir.name, 'conf.yml')
        with open(conf, 'w') as f:
            f.write('a: b\nc: 3')
        self.assertEqual({'a': 'b', 'c': 3}, _load_config(conf))

    def test_cwd(self):
        cwd = os.getcwd()
        file = 'test.yml'
        self.assertEqual(os.path.join(cwd, file),
                         _find_file(file, verify_later=True))

    def test_abspath(self):
        path = '/etc/config.yml'
        self.assertEqual(path,
                         _find_file(path, verify_later=True))

    def test_verify(self):
        path = os.path.join(self.tmpdir.name, 'does-not-exist.yml')
        self.assertIsNone(_find_file(path))


class TestGoodConf(TestCase):
    def test_set_values(self):
        c = GoodConf()
        c.define_values(Value('a'), Value('c', default=4))
        c.set_values({'a': 'b'})
        self.assertEqual(c.a, 'b')
        self.assertEqual(c.c, 4)

    def test_initial(self):
        c = GoodConf()
        c.define_values(Value('a', initial=lambda: True))
        self.assertEqual(c.get_initial(), {'a': True})

    def test_dump_json(self):
        c = GoodConf()
        c.define_values(Value('a', initial=lambda: True))
        self.assertEqual(c.generate_json(), '{\n  "a": true\n}')

    def test_dump_yaml(self):
        c = GoodConf(description="Configuration for My App")
        c.define_values(Value('a', help="this is a"))
        self.assertEqual(c.generate_yaml(), dedent("""
            # 
            # Configuration for My App
            # 
            
            # this is a
            a: ''
            """).lstrip())

    def test_generate_markdown(self):
        description = "Configuration for My App"
        help_ = "this is a"
        c = GoodConf(description=description)
        c.define_values(Value('a', help=help_, default=5),
                        Value('b', required=True))
        mkdn = c.generate_markdown()
        # Not sure on final format, just do some basic smoke tests
        self.assertIn(description, mkdn)
        self.assertIn(help_, mkdn)


class TestGoodConfFiles(TestCase):
    def setUp(self):
        self.tmpdir = TemporaryDirectory()

    def tearDown(self):
        self.tmpdir.cleanup()

    @mock.patch('goodconf._load_config')
    def test_conf_env_var(self, mocked_load_config):
        path = Path(self.tmpdir.name) / 'myapp.json'
        path.touch()
        path = str(path)
        c = GoodConf(file_env_var='CONF')
        with env_var('CONF', path):
            c.load()
        mocked_load_config.assert_called_once_with(path)

    @mock.patch('goodconf.GoodConf.set_values')
    def test_all_env_vars(self, mocked_set_values):
        c = GoodConf()
        c.load()
        mocked_set_values.assert_called_once_with({})

    @mock.patch('goodconf._load_config')
    def test_provided_file(self, mocked_load_config):
        path = Path(self.tmpdir.name) / 'myapp.json'
        path.touch()
        path = str(path)
        c = GoodConf()
        c.load(path)
        mocked_load_config.assert_called_once_with(path)

    @mock.patch('goodconf._load_config')
    def test_default_files(self, mocked_load_config):
        path = Path(self.tmpdir.name) / 'myapp.json'
        path.touch()
        path = str(path)
        bad_path = str(Path(self.tmpdir.name) / 'does-not-exist.json')
        c = GoodConf(default_files=[bad_path, path])
        c.load()
        mocked_load_config.assert_called_once_with(path)


def skip_if_no_django(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            import django
            return f(*args, **kwargs)
        except ImportError:
            raise SkipTest("Django is not installed")
    return wrapper



class TestDjango(TestCase):

    @skip_if_no_django
    @mock.patch('django.core.management.execute_from_command_line')
    @mock.patch('goodconf._load_config')
    def test_mgmt_command(self, mocked_load_config, mocked_dj_execute):
        c = GoodConf()
        path = '/etc/config.yml'
        dj_args = ['manage.py', 'diffsettings', '-v', '2']
        execute_from_command_line_with_config(c, dj_args + ['-c', path])
        mocked_load_config.assert_called_once_with('/etc/config.yml')
        mocked_dj_execute.assert_called_once_with(dj_args)

    @skip_if_no_django
    @mock.patch('goodconf._load_config')
    @mock.patch('sys.stdout', new_callable=StringIO)
    @mock.patch('sys.exit')
    def test_help(self, mocked_sysexit, mocked_stdout, mocked_load_config):
        path = '/etc/config.yml'
        c = GoodConf(file_env_var='MYAPP_CONF',
                     default_files=['/etc/myapp.json'])
        execute_from_command_line_with_config(c, [
            'manage.py', 'diffsettings', '-c', path,
            '--settings', 'tests.test_goodconf', '-h'])
        mocked_load_config.assert_called_once_with('/etc/config.yml')
        output = mocked_stdout.getvalue()
        self.assertIn('-c FILE, --config FILE', output)
        self.assertIn('MYAPP_CONF', output)
        self.assertIn('/etc/myapp.json', output)


# This doubles as a Django settings file for the tests
SECRET_KEY = 'abc'



import argparse
from typing import List

from goodconf import GoodConf
from .argparse import argparser_add_argument


def _monkeypatch_commandparser(config: GoodConf):
    """Adds -c, --config to Django BaseCommand"""
    from django.core.management.base import BaseCommand
    default_parser = BaseCommand.create_parser

    def patched_parser(self, prog_name, subcommand):
        parser = default_parser(self, prog_name, subcommand)
        argparser_add_argument(parser, config)
        return parser
    BaseCommand.create_parser = patched_parser


def load_config_from_cli(config: GoodConf, argv: List[str]) -> List[str]:
    """Loads config, checking CLI arguments for a config file"""
    _monkeypatch_commandparser(config)

    parser = argparse.ArgumentParser(add_help=False)
    argparser_add_argument(parser, config)

    config_arg, default_args = parser.parse_known_args(argv)
    config.load(config_arg.config)
    return default_args


def execute_from_command_line_with_config(config: GoodConf, argv: List[str]):
    """Load's config then runs Django's execute_from_command_line"""
    args = load_config_from_cli(config, argv)
    from django.core.management import execute_from_command_line
    execute_from_command_line(args)
